package bin.se.alanif.alanide.editors.formatters;

import java.io.InputStream;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Map;
import java.util.HashMap;

class Token {
	public int kind;    // token kind
	public int pos;     // token position in bytes in the source text (starting at 0)
	public int charPos; // token position in characters in the source text (starting at 0)
	public int col;     // token column (starting at 1)
	public int line;    // token line (starting at 1)
	public String val;  // token value
	public Token next;  // ML 2005-03-11 Peek tokens are kept in linked list
}

//-----------------------------------------------------------------------------------
// Buffer
//-----------------------------------------------------------------------------------
class Buffer {
	// This Buffer supports the following cases:
	// 1) seekable stream (file)
	//    a) whole stream in buffer
	//    b) part of stream in buffer
	// 2) non seekable stream (network, console)

	public static final int EOF = Character.MAX_VALUE + 1;
	private static final int MIN_BUFFER_LENGTH = 1024; // 1KB
	private static final int MAX_BUFFER_LENGTH = MIN_BUFFER_LENGTH * 64; // 64KB
	private byte[] buf;   // input buffer
	private int bufStart; // position of first byte in buffer relative to input stream
	private int bufLen;   // length of buffer
	private int fileLen;  // length of input stream (may change if stream is no file)
	private int bufPos;      // current position in buffer
	private RandomAccessFile file; // input stream (seekable)
	private InputStream stream; // growing input stream (e.g.: console, network)

	public Buffer(InputStream s) {
		stream = s;
		fileLen = bufLen = bufStart = bufPos = 0;
		buf = new byte[MIN_BUFFER_LENGTH];
	}

	public Buffer(String fileName) {
		try {
			file = new RandomAccessFile(fileName, "r");
			fileLen = (int) file.length();
			bufLen = Math.min(fileLen, MAX_BUFFER_LENGTH);
			buf = new byte[bufLen];
			bufStart = Integer.MAX_VALUE; // nothing in buffer so far
			if (fileLen > 0) setPos(0); // setup buffer to position 0 (start)
			else bufPos = 0; // index 0 is already after the file, thus setPos(0) is invalid
			if (bufLen == fileLen) Close();
		} catch (IOException e) {
			throw new FatalError("Could not open file " + fileName);
		}
	}

	// don't use b after this call anymore
	// called in UTF8Buffer constructor
	protected Buffer(Buffer b) {
		buf = b.buf;
		bufStart = b.bufStart;
		bufLen = b.bufLen;
		fileLen = b.fileLen;
		bufPos = b.bufPos;
		file = b.file;
		stream = b.stream;
		// keep finalize from closing the file
		b.file = null;
	}

	protected void finalize() throws Throwable {
		super.finalize();
		Close();
	}

	protected void Close() {
		if (file != null) {
			try {
				file.close();
				file = null;
			} catch (IOException e) {
				throw new FatalError(e.getMessage());
			}
		}
	}

	public int Read() {
		if (bufPos < bufLen) {
			return buf[bufPos++] & 0xff;  // mask out sign bits
		} else if (getPos() < fileLen) {
			setPos(getPos());         // shift buffer start to pos
			return buf[bufPos++] & 0xff; // mask out sign bits
		} else if (stream != null && ReadNextStreamChunk() > 0) {
			return buf[bufPos++] & 0xff;  // mask out sign bits
		} else {
			return EOF;
		}
	}

	public int Peek() {
		int curPos = getPos();
		int ch = Read();
		setPos(curPos);
		return ch;
	}

	// beg .. begin, zero-based, inclusive, in byte
	// end .. end, zero-based, exclusive, in byte
	public String GetString(int beg, int end) {
		int len = 0;
		char[] buf = new char[end - beg];
		int oldPos = getPos();
		setPos(beg);
		while (getPos() < end) buf[len++] = (char) Read();
		setPos(oldPos);
		return new String(buf, 0, len);
	}

	public int getPos() {
		return bufPos + bufStart;
	}

	public void setPos(int value) {
		if (value >= fileLen && stream != null) {
			// Wanted position is after buffer and the stream
			// is not seek-able e.g. network or console,
			// thus we have to read the stream manually till
			// the wanted position is in sight.
			while (value >= fileLen && ReadNextStreamChunk() > 0);
		}

		if (value < 0 || value > fileLen) {
			throw new FatalError("buffer out of bounds access, position: " + value);
		}

		if (value >= bufStart && value < bufStart + bufLen) { // already in buffer
			bufPos = value - bufStart;
		} else if (file != null) { // must be swapped in
			try {
				file.seek(value);
				bufLen = file.read(buf);
				bufStart = value; bufPos = 0;
			} catch(IOException e) {
				throw new FatalError(e.getMessage());
			}
		} else {
			// set the position to the end of the file, Pos will return fileLen.
			bufPos = fileLen - bufStart;
		}
	}
	
	// Read the next chunk of bytes from the stream, increases the buffer
	// if needed and updates the fields fileLen and bufLen.
	// Returns the number of bytes read.
	private int ReadNextStreamChunk() {
		int free = buf.length - bufLen;
		if (free == 0) {
			// in the case of a growing input stream
			// we can neither seek in the stream, nor can we
			// foresee the maximum length, thus we must adapt
			// the buffer size on demand.
			byte[] newBuf = new byte[bufLen * 2];
			System.arraycopy(buf, 0, newBuf, 0, bufLen);
			buf = newBuf;
			free = bufLen;
		}
		
		int read;
		try { read = stream.read(buf, bufLen, free); }
		catch (IOException ioex) { throw new FatalError(ioex.getMessage()); }
		
		if (read > 0) {
			fileLen = bufLen = (bufLen + read);
			return read;
		}
		// end of stream reached
		return 0;
	}
}

//-----------------------------------------------------------------------------------
// UTF8Buffer
//-----------------------------------------------------------------------------------
class UTF8Buffer extends Buffer {
	UTF8Buffer(Buffer b) { super(b); }

	public int Read() {
		int ch;
		do {
			ch = super.Read();
			// until we find a utf8 start (0xxxxxxx or 11xxxxxx)
		} while ((ch >= 128) && ((ch & 0xC0) != 0xC0) && (ch != EOF));
		if (ch < 128 || ch == EOF) {
			// nothing to do, first 127 chars are the same in ascii and utf8
			// 0xxxxxxx or end of file character
		} else if ((ch & 0xF0) == 0xF0) {
			// 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
			int c1 = ch & 0x07; ch = super.Read();
			int c2 = ch & 0x3F; ch = super.Read();
			int c3 = ch & 0x3F; ch = super.Read();
			int c4 = ch & 0x3F;
			ch = (((((c1 << 6) | c2) << 6) | c3) << 6) | c4;
		} else if ((ch & 0xE0) == 0xE0) {
			// 1110xxxx 10xxxxxx 10xxxxxx
			int c1 = ch & 0x0F; ch = super.Read();
			int c2 = ch & 0x3F; ch = super.Read();
			int c3 = ch & 0x3F;
			ch = (((c1 << 6) | c2) << 6) | c3;
		} else if ((ch & 0xC0) == 0xC0) {
			// 110xxxxx 10xxxxxx
			int c1 = ch & 0x1F; ch = super.Read();
			int c2 = ch & 0x3F;
			ch = (c1 << 6) | c2;
		}
		return ch;
	}
}

//-----------------------------------------------------------------------------------
// StartStates  -- maps characters to start states of tokens
//-----------------------------------------------------------------------------------
class StartStates {
	private static class Elem {
		public int key, val;
		public Elem next;
		public Elem(int key, int val) { this.key = key; this.val = val; }
	}

	private Elem[] tab = new Elem[128];

	public void set(int key, int val) {
		Elem e = new Elem(key, val);
		int k = key % 128;
		e.next = tab[k]; tab[k] = e;
	}

	public int state(int key) {
		Elem e = tab[key % 128];
		while (e != null && e.key != key) e = e.next;
		return e == null ? 0: e.val;
	}
}

//-----------------------------------------------------------------------------------
// Scanner
//-----------------------------------------------------------------------------------
public class Scanner {
	static final char EOL = '\n';
	static final int  eofSym = 0;
	static final int maxT = 146;
	static final int noSym = 146;
	char valCh;       // current input character (for token.val)

	public Buffer buffer; // scanner buffer

	Token t;           // current token
	int ch;            // current input character
	int pos;           // byte position of current character
	int charPos;       // position by unicode characters starting with 0
	int col;           // column number of current character
	int line;          // line number of current character
	int oldEols;       // EOLs that appeared in a comment;
	static final StartStates start; // maps initial token character to start state
	static final Map literals;      // maps literal strings to literal kinds

	Token tokens;      // list of tokens already peeked (first token is a dummy)
	Token pt;          // current peek token
	
	char[] tval = new char[16]; // token text used in NextToken(), dynamically enlarged
	int tlen;          // length of current token


	static {
		start = new StartStates();
		literals = new HashMap();
		for (int i = 36; i <= 36; ++i) start.set(i, 1);
		for (int i = 95; i <= 95; ++i) start.set(i, 1);
		for (int i = 97; i <= 122; ++i) start.set(i, 1);
		for (int i = 224; i <= 246; ++i) start.set(i, 1);
		for (int i = 248; i <= 254; ++i) start.set(i, 1);
		for (int i = 48; i <= 57; ++i) start.set(i, 4);
		start.set(39, 2); 
		start.set(34, 5); 
		start.set(46, 7); 
		start.set(45, 8); 
		start.set(123, 9); 
		start.set(125, 10); 
		start.set(44, 11); 
		start.set(61, 24); 
		start.set(58, 12); 
		start.set(40, 13); 
		start.set(41, 14); 
		start.set(42, 25); 
		start.set(33, 26); 
		start.set(43, 18); 
		start.set(47, 19); 
		start.set(60, 27); 
		start.set(62, 28); 
		start.set(Buffer.EOF, -1);
		literals.put("opaque", new Integer(4));
		literals.put("container", new Integer(5));
		literals.put("not", new Integer(6));
		literals.put("option", new Integer(7));
		literals.put("options", new Integer(8));
		literals.put("prompt", new Integer(10));
		literals.put("import", new Integer(11));
		literals.put("synonyms", new Integer(16));
		literals.put("message", new Integer(18));
		literals.put("syntax", new Integer(20));
		literals.put("and", new Integer(27));
		literals.put("isa", new Integer(28));
		literals.put("else", new Integer(29));
		literals.put("where", new Integer(30));
		literals.put("verb", new Integer(31));
		literals.put("when", new Integer(32));
		literals.put("end", new Integer(33));
		literals.put("check", new Integer(34));
		literals.put("does", new Integer(35));
		literals.put("every", new Integer(36));
		literals.put("add", new Integer(37));
		literals.put("to", new Integer(38));
		literals.put("the", new Integer(39));
		literals.put("definite", new Integer(40));
		literals.put("indefinite", new Integer(41));
		literals.put("negative", new Integer(42));
		literals.put("exit", new Integer(43));
		literals.put("is", new Integer(44));
		literals.put("are", new Integer(45));
		literals.put("has", new Integer(46));
		literals.put("can", new Integer(47));
		literals.put("description", new Integer(48));
		literals.put("article", new Integer(49));
		literals.put("form", new Integer(50));
		literals.put("entered", new Integer(51));
		literals.put("initialize", new Integer(52));
		literals.put("mentioned", new Integer(53));
		literals.put("name", new Integer(54));
		literals.put("pronoun", new Integer(55));
		literals.put("with", new Integer(56));
		literals.put("taking", new Integer(57));
		literals.put("limits", new Integer(58));
		literals.put("then", new Integer(59));
		literals.put("count", new Integer(60));
		literals.put("header", new Integer(61));
		literals.put("extract", new Integer(62));
		literals.put("event", new Integer(63));
		literals.put("script", new Integer(64));
		literals.put("step", new Integer(65));
		literals.put("after", new Integer(66));
		literals.put("wait", new Integer(67));
		literals.put("until", new Integer(68));
		literals.put("start", new Integer(70));
		literals.put("describe", new Integer(71));
		literals.put("say", new Integer(72));
		literals.put("list", new Integer(73));
		literals.put("show", new Integer(74));
		literals.put("play", new Integer(75));
		literals.put("style", new Integer(76));
		literals.put("an", new Integer(77));
		literals.put("it", new Integer(78));
		literals.put("no", new Integer(79));
		literals.put("empty", new Integer(80));
		literals.put("locate", new Integer(81));
		literals.put("include", new Integer(82));
		literals.put("in", new Integer(83));
		literals.put("exclude", new Integer(84));
		literals.put("from", new Integer(85));
		literals.put("cancel", new Integer(86));
		literals.put("schedule", new Integer(87));
		literals.put("make", new Integer(88));
		literals.put("strip", new Integer(89));
		literals.put("increase", new Integer(90));
		literals.put("by", new Integer(91));
		literals.put("decrease", new Integer(92));
		literals.put("set", new Integer(93));
		literals.put("first", new Integer(94));
		literals.put("last", new Integer(95));
		literals.put("word", new Integer(96));
		literals.put("words", new Integer(97));
		literals.put("character", new Integer(98));
		literals.put("characters", new Integer(99));
		literals.put("into", new Integer(100));
		literals.put("if", new Integer(101));
		literals.put("elsif", new Integer(102));
		literals.put("depending", new Integer(103));
		literals.put("on", new Integer(104));
		literals.put("depend", new Integer(105));
		literals.put("do", new Integer(106));
		literals.put("between", new Integer(107));
		literals.put("for", new Integer(108));
		literals.put("each", new Integer(109));
		literals.put("stop", new Integer(110));
		literals.put("use", new Integer(111));
		literals.put("quit", new Integer(112));
		literals.put("look", new Integer(113));
		literals.put("save", new Integer(114));
		literals.put("restore", new Integer(115));
		literals.put("restart", new Integer(116));
		literals.put("score", new Integer(117));
		literals.put("visits", new Integer(118));
		literals.put("system", new Integer(119));
		literals.put("or", new Integer(120));
		literals.put("contains", new Integer(121));
		literals.put("random", new Integer(122));
		literals.put("directly", new Integer(123));
		literals.put("of", new Integer(124));
		literals.put("max", new Integer(125));
		literals.put("min", new Integer(126));
		literals.put("sum", new Integer(127));
		literals.put("this", new Integer(128));
		literals.put("current", new Integer(129));
		literals.put("actor", new Integer(130));
		literals.put("location", new Integer(131));
		literals.put("here", new Integer(132));
		literals.put("nearby", new Integer(133));
		literals.put("at", new Integer(134));
		literals.put("near", new Integer(135));
		literals.put("before", new Integer(144));
		literals.put("only", new Integer(145));

	}
	
	public Scanner (String fileName) {
		buffer = new Buffer(fileName);
		Init();
	}
	
	public Scanner(InputStream s) {
		buffer = new Buffer(s);
		Init();
	}
	
	void Init () {
		pos = -1; line = 1; col = 0; charPos = -1;
		oldEols = 0;
		NextCh();
		if (ch == 0xEF) { // check optional byte order mark for UTF-8
			NextCh(); int ch1 = ch;
			NextCh(); int ch2 = ch;
			if (ch1 != 0xBB || ch2 != 0xBF) {
				throw new FatalError("Illegal byte order mark at start of file");
			}
			buffer = new UTF8Buffer(buffer); col = 0; charPos = -1;
			NextCh();
		}
		pt = tokens = new Token();  // first token is a dummy
	}
	
	void NextCh() {
		if (oldEols > 0) { ch = EOL; oldEols--; }
		else {
			pos = buffer.getPos();
			// buffer reads unicode chars, if UTF8 has been detected
			ch = buffer.Read(); col++; charPos++;
			// replace isolated '\r' by '\n' in order to make
			// eol handling uniform across Windows, Unix and Mac
			if (ch == '\r' && buffer.Peek() != '\n') ch = EOL;
			if (ch == EOL) { line++; col = 0; }
		}
		if (ch != Buffer.EOF) {
			valCh = (char) ch;
			ch = Character.toLowerCase(ch);
		}

	}
	
	void AddCh() {
		if (tlen >= tval.length) {
			char[] newBuf = new char[2 * tval.length];
			System.arraycopy(tval, 0, newBuf, 0, tval.length);
			tval = newBuf;
		}
		if (ch != Buffer.EOF) {
			tval[tlen++] = valCh; 

			NextCh();
		}

	}
	

	boolean Comment0() {
		int level = 1, pos0 = pos, line0 = line, col0 = col, charPos0 = charPos;
		NextCh();
		if (ch == '-') {
			NextCh();
			for(;;) {
				if (ch == 10) {
					level--;
					if (level == 0) { oldEols = line - line0; NextCh(); return true; }
					NextCh();
				} else if (ch == Buffer.EOF) return false;
				else NextCh();
			}
		} else {
			buffer.setPos(pos0); NextCh(); line = line0; col = col0; charPos = charPos0;
		}
		return false;
	}


	void CheckLiteral() {
		String val = t.val;
		val = val.toLowerCase();

		Object kind = literals.get(val);
		if (kind != null) {
			t.kind = ((Integer) kind).intValue();
		}
	}

	Token NextToken() {
		while (ch == ' ' ||
			ch >= 9 && ch <= 10 || ch == 13
		) NextCh();
		if (ch == '-' && Comment0()) return NextToken();
		int recKind = noSym;
		int recEnd = pos;
		t = new Token();
		t.pos = pos; t.col = col; t.line = line; t.charPos = charPos;
		int state = start.state(ch);
		tlen = 0; AddCh();

		loop: for (;;) {
			switch (state) {
				case -1: { t.kind = eofSym; break loop; } // NextCh already done 
				case 0: {
					if (recKind != noSym) {
						tlen = recEnd - t.pos;
						SetScannerBehindT();
					}
					t.kind = recKind; break loop;
				} // NextCh already done
				case 1:
					recEnd = pos; recKind = 1;
					if (ch == '$' || ch >= '0' && ch <= '9' || ch == '_' || ch >= 'a' && ch <= 'z' || ch >= 224 && ch <= 246 || ch >= 248 && ch <= 254) {AddCh(); state = 1; break;}
					else {t.kind = 1; t.val = new String(tval, 0, tlen); CheckLiteral(); return t;}
				case 2:
					if (ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '&' || ch >= '(' && ch <= 65535) {AddCh(); state = 2; break;}
					else if (ch == 39) {AddCh(); state = 3; break;}
					else {state = 0; break;}
				case 3:
					{t.kind = 1; t.val = new String(tval, 0, tlen); CheckLiteral(); return t;}
				case 4:
					recEnd = pos; recKind = 2;
					if (ch >= '0' && ch <= '9') {AddCh(); state = 4; break;}
					else {t.kind = 2; break loop;}
				case 5:
					if (ch <= '!' || ch >= '#' && ch <= 65535) {AddCh(); state = 5; break;}
					else if (ch == '"') {AddCh(); state = 6; break;}
					else {state = 0; break;}
				case 6:
					recEnd = pos; recKind = 3;
					if (ch == '"') {AddCh(); state = 5; break;}
					else {t.kind = 3; break loop;}
				case 7:
					{t.kind = 9; break loop;}
				case 8:
					{t.kind = 12; break loop;}
				case 9:
					{t.kind = 13; break loop;}
				case 10:
					{t.kind = 14; break loop;}
				case 11:
					{t.kind = 15; break loop;}
				case 12:
					{t.kind = 19; break loop;}
				case 13:
					{t.kind = 21; break loop;}
				case 14:
					{t.kind = 22; break loop;}
				case 15:
					{t.kind = 25; break loop;}
				case 16:
					{t.kind = 26; break loop;}
				case 17:
					{t.kind = 69; break loop;}
				case 18:
					{t.kind = 136; break loop;}
				case 19:
					{t.kind = 137; break loop;}
				case 20:
					{t.kind = 138; break loop;}
				case 21:
					{t.kind = 139; break loop;}
				case 22:
					{t.kind = 140; break loop;}
				case 23:
					{t.kind = 141; break loop;}
				case 24:
					recEnd = pos; recKind = 17;
					if (ch == '>') {AddCh(); state = 17; break;}
					else if (ch == '=') {AddCh(); state = 21; break;}
					else {t.kind = 17; break loop;}
				case 25:
					recEnd = pos; recKind = 23;
					if (ch == '!') {AddCh(); state = 15; break;}
					else {t.kind = 23; break loop;}
				case 26:
					recEnd = pos; recKind = 24;
					if (ch == '*') {AddCh(); state = 16; break;}
					else {t.kind = 24; break loop;}
				case 27:
					recEnd = pos; recKind = 143;
					if (ch == '>') {AddCh(); state = 20; break;}
					else if (ch == '=') {AddCh(); state = 23; break;}
					else {t.kind = 143; break loop;}
				case 28:
					recEnd = pos; recKind = 142;
					if (ch == '=') {AddCh(); state = 22; break;}
					else {t.kind = 142; break loop;}

			}
		}
		t.val = new String(tval, 0, tlen);
		return t;
	}
	
	private void SetScannerBehindT() {
		buffer.setPos(t.pos);
		NextCh();
		line = t.line; col = t.col; charPos = t.charPos;
		for (int i = 0; i < tlen; i++) NextCh();
	}
	
	// get the next token (possibly a token already seen during peeking)
	public Token Scan () {
		if (tokens.next == null) {
			return NextToken();
		} else {
			pt = tokens = tokens.next;
			return tokens;
		}
	}

	// get the next token, ignore pragmas
	public Token Peek () {
		do {
			if (pt.next == null) {
				pt.next = NextToken();
			}
			pt = pt.next;
		} while (pt.kind > maxT); // skip pragmas

		return pt;
	}

	// make sure that peeking starts at current scan position
	public void ResetPeek () { pt = tokens; }

} // end Scanner
